/**************************************************************************
 *
 * FILE  cbm_mem.c
 * Copyright (c) 2007, 2012, 2018 Daniel Kahlin
 * Written by Daniel Kahlin <daniel@kahlin.net>
 *
 * DESCRIPTION
 *   cbm memory handling
 *
 ******/
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "cbm_mem.h"
#include "global.h"

int is_defined_addr(cbm_ptr_t ad)
{
    return ad != CBM_PTR_UNDEFINED;
}

cbm_ptr_t min_addr(cbm_ptr_t ad1, cbm_ptr_t ad2)
{
    if (!is_defined_addr(ad1) || ad1 > ad2) {
	return ad2;
    }
    return ad1;
}

cbm_ptr_t max_addr(cbm_ptr_t ad1, cbm_ptr_t ad2)
{
    if (!is_defined_addr(ad1) || ad1 < ad2) {
	return ad2;
    }
    return ad1;
}

int is_defined_range(cbm_range_t rng)
{
    return is_defined_addr(rng.sa) && is_defined_addr(rng.ea);
}

void update_range(cbm_range_t *rng, cbm_ptr_t ad)
{
    if (ad < rng->sa || !is_defined_addr(rng->sa)) {
	rng->sa = ad;
    }
    if (rng->ea < ad + 1 || !is_defined_addr(rng->ea)) {
	rng->ea = ad + 1;
    }
}

cbm_range_t max_range(cbm_range_t rng1, cbm_range_t rng2)
{
    cbm_range_t rng;

    rng.sa = min_addr(rng1.sa, rng2.sa);
    rng.ea = max_addr(rng1.ea, rng2.ea);

    return rng;
}


cbm_mem_t *create_mem(void)
{
    cbm_mem_t *cm;
    int size = RAM_SIZE;

    cm = malloc(sizeof(cbm_mem_t));
    if (!cm) {
	panic("couldn't allocate mem");
    }

    cm->ram = malloc(size);
    cm->usemap = malloc(size);
    if ( !(cm->ram && cm->usemap) ) {
	destroy_mem(cm);
	panic("couldn't allocate mem");
    }

    clear_mem(cm, 0, size);
    cm->size = size;

    return cm;
}

void destroy_mem(cbm_mem_t *cm)
{
    if (cm) {
	free(cm->ram);
	free(cm->usemap);
	free(cm);
    }
}


static cbm_ptr_t wrap_addr(cbm_mem_t *cm, cbm_ptr_t ad)
{
    return ad & (cm->size - 1);
}

static int range_len(cbm_mem_t *cm, cbm_ptr_t sa, cbm_ptr_t ea)
{
#if 0
    return wrap_addr(cm, ea - sa);
#else
    if (ea > sa) {
        return ea - sa;
    } else {
        return wrap_addr(cm, ea - sa);
    }
#endif
}


uint8_t get_byte(cbm_mem_t *cm, cbm_ptr_t ad)
{
    if (cm->usemap[wrap_addr(cm, ad)]) {
	return cm->ram[wrap_addr(cm, ad)];
    }

    return 0x00;
}

uint8_t get_byte_defined(cbm_mem_t *cm, cbm_ptr_t ad)
{
    return cm->usemap[wrap_addr(cm, ad)];
}


void set_byte(cbm_mem_t *cm, cbm_ptr_t ad, uint8_t val)
{
    cm->ram[wrap_addr(cm, ad)] = val;
    cm->usemap[wrap_addr(cm, ad)] = 1;
}

void unset_byte(cbm_mem_t *cm, cbm_ptr_t ad)
{
    cm->ram[wrap_addr(cm, ad)] = 0x00;
    cm->usemap[wrap_addr(cm, ad)] = 0;
}

uint16_t get_word(cbm_mem_t *cm, cbm_ptr_t ad)
{
    return get_byte(cm, ad) | (get_byte(cm, ad + 1) << 8);
}

void set_word(cbm_mem_t *cm, cbm_ptr_t ad, uint16_t val)
{
    set_byte(cm, ad, val & 0xff);
    set_byte(cm, ad + 1, val >> 8);
}

cbm_ptr_t get_low_addr(cbm_mem_t *cm)
{
    cbm_ptr_t ad;
    for (ad = 0; ad < cm->size; ad++) {
	if (get_byte_defined(cm, ad)) {
	    return ad;
	}
    }

    return CBM_PTR_UNDEFINED;
}

cbm_ptr_t get_high_addr(cbm_mem_t *cm)
{
    cbm_ptr_t ad;

    ad = cm->size;
    do {
	ad--;
	if (get_byte_defined(cm, ad)) {
	    return ad+1;
	}
    } while (ad > 0);

    return CBM_PTR_UNDEFINED;
}

cbm_range_t get_range(cbm_mem_t *cm)
{
    cbm_range_t rng;

    rng.sa = get_low_addr(cm);
    rng.ea = get_high_addr(cm);

    return rng;
}


void clear_mem(cbm_mem_t *cm, cbm_ptr_t sa, cbm_ptr_t ea)
{
    int n = range_len(cm, sa, ea);

    int i;
    for (i = 0; i < n; i++) {
	unset_byte(cm, sa + i);
    }
}

void copy_mem(cbm_mem_t *cm, cbm_ptr_t sa, cbm_ptr_t ea, cbm_ptr_t da)
{
    copy_mem2(cm, sa, ea, cm, da);
}

void copy_mem2(cbm_mem_t *scm, cbm_ptr_t sa, cbm_ptr_t ea, cbm_mem_t *dcm, cbm_ptr_t da)
{
    int n = range_len(scm, sa, ea);

    int i;
    if (da < sa) {
	for (i = 0; i < n; i++) {
            set_byte(dcm, da + i, get_byte(scm, sa + i));
	}
    } else {
	for (i = n - 1; i >= 0; --i) {
            set_byte(dcm, da + i, get_byte(scm, sa + i));
	}
    }
}

void insert_mem(uint8_t *src, int len, cbm_mem_t *dcm, cbm_ptr_t da)
{
    int i;
    for (i = 0; i < len; i++) {
	set_byte(dcm, da + i, src[i]);
    }
}

void load_mem(cbm_mem_t *cm, const char *name, cbm_ptr_t la, int offs, int len, enum mode_t mode, cbm_ptr_t *aptr, int *lptr)
{
    int c;
    FILE *fp;
    int lrd;
    cbm_ptr_t ad;
    const char *print_name;

    if (!strcmp(name, "-")) {
	print_name = "<stdin>";
	fp = stdin;
    } else {
	print_name = name;
	fp = fopen(name, "rb");
	if (!fp) {
	    panic("%s: %s", print_name, strerror(errno));
	}
    }

    switch (mode) {
    case MODE_NORMAL:
	/* get load address */
	la = fgetc(fp) + (fgetc(fp) << 8);
	break;
    case MODE_NEWADDR:
	/* skip load address */
	fgetc(fp);
	fgetc(fp);
	break;
    case MODE_RAW:
	/* no load address */
	break;
    default:
	break;
    }

    /* skip offset if any */
    if (offs > 0) {
	while ( c = fgetc(fp), c != EOF ) {
	    if (mode == MODE_NORMAL) {
		/* ugly but working */
		la++;
	    }
	    offs--;
	    if (offs == 0) {
		break;
	    }
	}
    }

    /* load file body */
    ad = la;
    lrd = 0;
    if (len != 0) {
	while ( c = fgetc(fp), c != EOF ) {
	    set_byte(cm, ad, c);
	    ad++;
	    lrd++;
	    /* if a max len is specified, then terminate when it has been
	       reached. */
	    if (len > 0 && lrd >= len) {
		break;
	    }
	}
    }

    if (ferror(fp)) {
	panic("%s: %s", print_name, strerror(errno));
    }

    if (fp != stdin) {
	c = fclose(fp);
	if (c == EOF) {
	    panic("%s: %s", print_name, strerror(errno));
	}
    }
    if (verbose_g) {
	printf("read '%s' $%04X-$%04X.\n", print_name, la, ad);
    }

    if (aptr) {
	*aptr = la;
    }
    if (lptr) {
	*lptr = lrd;
    }
}


void save_mem(cbm_mem_t *cm, const char *name, cbm_ptr_t sa, cbm_ptr_t ea, cbm_ptr_t la, enum mode_t mode)
{
    int c;
    FILE *fp;
    cbm_ptr_t ad;
    const char *print_name;

    /* write destination file */
    if (!strcmp(name, "-")) {
	print_name = "<stdout>";
	fp = stdout;
    } else {
	print_name = name;
	fp = fopen(name, "wb");
	if (!fp) {
	    panic("%s: %s", print_name, strerror(errno));
	}
    }

    switch (mode) {
    case MODE_NORMAL:
	/* set load address to starting address */
	/* write load address */
	fputc(sa & 0xff, fp);
	fputc(sa >> 8, fp);
	break;
    case MODE_NEWADDR:
	/* use alternate load address (in la) */
	fputc(la & 0xff, fp);
	fputc(la >> 8, fp);
	break;
    case MODE_RAW:
	/* no load address */
	break;
    default:
	break;
    }

    /* write file body */
    ad = sa;
    while (ad != ea) {
	uint8_t c = get_byte(cm, ad);
	fputc(c, fp);
	ad++;
    }

    if (ferror(fp)) {
	panic("%s: %s", print_name, strerror(errno));
    }

    if (fp != stdout) {
	c = fclose(fp);
	if (c == EOF) {
	    panic("%s: %s", print_name, strerror(errno));
	}
    }
    if (verbose_g) {
	printf("wrote '%s' $%04X-$%04X.\n", print_name, sa, ea);
    }
}

/* eof */
