/**************************************************************************
 *
 * FILE  cdif.c
 * Copyright (c) 2013, 2018 Daniel Kahlin
 * Written by Daniel Kahlin <daniel@kahlin.net>
 *
 * DESCRIPTION
 *   cbm file binary diff
 *
 ******/
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "cbm_mem.h"
#include "global.h"
#include "params.h"

#define PROGRAM "cdif"

const char program_g[] = PROGRAM;

typedef struct {
    char *outname;
} Options;

static Options opt;


int main(int argc, char *argv[])
{
    int c;
    char *srcname1;
    char *srcname2;

    /* defaults */
    srcname1 = "";
    srcname2 = "";
    verbose_g = 0;
    debug_g = 0;
    opt.outname = "a.out";

    /*
     * scan for valid options
     */
    while ( c = getopt(argc, argv, "o:dvVh"), c != EOF ) {
        switch (c) {

	/* a missing parameter */
	case ':':
	/* an illegal option */
	case '?':
	    exit(1);

	/* set verbose */
	case 'v':
	    verbose_g = 1;
	    break;

	/* set debug */
	case 'd':
	    debug_g = 1;
	    break;

	/* print version */
	case 'V':
	    fprintf(stdout, PROGRAM " (" PACKAGE ") " VERSION "\n");
	    exit(0);

	/* print help */
	case 'h':
	    fprintf(stderr,
PROGRAM " (" PACKAGE ") " VERSION ": cbm binary diff\n"
"Copyright (c) 2013, 2018 Daniel Kahlin <daniel@kahlin.net>\n"
"Written by Daniel Kahlin <daniel@kahlin.net>\n"
"\n"
"usage: " PROGRAM " [-o<name>][-v][-d][-h][-V] <file>...\n"
"\n"
"Valid options:\n"
"    -o              output file\n"
"    -v              be verbose\n"
"    -d              debugging output\n"
"    -h              displays this help text\n"
"    -V              output program version\n"
	    );
	    exit(0);

	/* set output file */
	case 'o':
	    opt.outname = optarg;
	    break;

	/* default behavior */
	default:
	    break;
	}
    }

    /*
     * optind now points at the first non option argument
     */
    int num_files = argc - optind;
    if (num_files < 2) {
	panic("too few arguments");
    }
    srcname1 = argv[optind++];
    srcname2 = argv[optind];

    cbm_mem_t *cm1 = create_mem();
    cbm_mem_t *cm2 = create_mem();

    loadfile_t f1, f2;
    f1 = parse_loadname(srcname1);
    load_mem(cm1, f1.name, f1.la, f1.offs, f1.len, f1.mode, 0, 0);

    f2 = parse_loadname(srcname2);
    load_mem(cm2, f2.name, f2.la, f2.offs, f2.len, f2.mode, 0, 0);

    cbm_ptr_t sa, ea;
    sa = min_addr(get_low_addr(cm1), get_low_addr(cm2));
    ea = max_addr(get_high_addr(cm1), get_high_addr(cm2));


    int is_different = 0;
    cbm_ptr_t ad;
    for (ad = sa; ad < ea; ad++) {
	int v1 = -1, v2 = -1;
	if (get_byte_defined(cm1, ad)) {
	    v1 = get_byte(cm1, ad);
	}
	if (get_byte_defined(cm2, ad)) {
	    v2 = get_byte(cm2, ad);
	}

	if (v1 != v2) {
	    printf("%04X: ", ad);
	    if (v1 < 0) {
		printf("-- ");
	    } else {
		printf("%02X ", v1);
	    }
	    if (v2 < 0) {
		printf("--\n");
	    } else {
		printf("%02X\n", v2);
	    }
	    is_different = 1;
	}


    }


    destroy_mem(cm2);
    destroy_mem(cm1);

    exit(is_different);
}

/* eof */
