/**************************************************************************
 *
 * FILE  cgen.c
 * Copyright (c) 2012, 2017, 2018 Daniel Kahlin
 * Written by Daniel Kahlin <daniel@kahlin.net>
 *
 * DESCRIPTION
 *   cbm test file generator
 *
 ******/
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <time.h>

#include "cbm_mem.h"
#include "checkers.h"
#include "global.h"
#include "params.h"

/* hack to fix random for windows compiles */
/* _WIN32 is defined on both 32-bit and 64-bit windows compiles */
#ifdef _WIN32
#define srandom(x) srand(x)
#define random(x) rand(x)
#endif


#define PROGRAM "cgen"

const char program_g[] = PROGRAM;

typedef struct {
    int check;
    char *outname;
    unsigned int seed;
    cbm_range_t rng;
    cbm_ptr_t load;
} Options;

static Options opt;


#define HAVE_SELF_SUM

static void generate(cbm_mem_t *cm, cbm_ptr_t sa, cbm_ptr_t ea, cbm_ptr_t load, unsigned int seed);

static void check(cbm_mem_t *cm);


int main(int argc, char *argv[])
{
    int c;

    /* defaults */
    verbose_g = 0;
    debug_g = 0;
    opt.check = 0;
    opt.outname = "a.out";
    opt.seed = time(0);
    opt.rng.sa = 0x0801;
    opt.rng.ea = 0xd000;
    opt.load = 0x0801;

    /*
     * scan for valid options
     */
    while ( c = getopt(argc, argv, "cr:j:S:o:dvVh"), c != EOF ) {
        switch (c) {

	/* a missing parameter */
	case ':':
	/* an illegal option */
	case '?':
	    exit(1);

	/* set verbose */
	case 'v':
	    verbose_g = 1;
	    break;

	/* set debug */
	case 'd':
	    debug_g = 1;
	    break;

	/* print version */
	case 'V':
	    fprintf(stdout, PROGRAM " (" PACKAGE ") " VERSION "\n");
	    exit(0);

	/* print help */
	case 'h':
	    fprintf(stderr,
PROGRAM " (" PACKAGE ") " VERSION ": cbm self-checking test file generator\n"
"Copyright (c) 2012, 2017, 2018 Daniel Kahlin <daniel@kahlin.net>\n"
"Written by Daniel Kahlin <daniel@kahlin.net>\n"
"\n"
"usage: " PROGRAM " [-r<range>][-j<addr>][-S<seed>][-o<name>][-v][-d][-h][-V]\n"
"       " PROGRAM " [-c][-v][-d] <file>\n"
"\n"
"Valid options:\n"
"    -c              check file integrity\n"
"    -r              range to generate\n"
"    -j              jump address\n"
"    -S              set fixed seed\n"
"    -o              output file\n"
"    -v              be verbose\n"
"    -d              debugging output\n"
"    -h              displays this help text\n"
"    -V              output program version\n"
	    );
	    exit(0);

	/* check file instead of generate */
	case 'c':
	    opt.check = 1;
	    break;

	/* set range */
	case 'r':
	    parse_range(optarg, &opt.rng);
	    break;

	/* set jump addr */
	case 'j':
	    opt.load = parse_value(optarg);
	    break;

	/* set seed */
	case 'S':
	    opt.seed = parse_value(optarg);
	    break;

	/* set output file */
	case 'o':
	    opt.outname = optarg;
	    break;

	/* default behavior */
	default:
	    break;
	}
    }

    /*
     * optind now points at the first non option argument
     */
    int num_files = argc - optind;
    char *srcname = "";

    if (!opt.check) {
	if (num_files > 0) {
	    panic("too many arguments");
	}
    } else {
	if (num_files < 1) {
	    panic("too few arguments");
	}
	srcname = argv[optind];
    }

    if (opt.check) {
	cbm_mem_t *cm = create_mem();
	loadfile_t f;

        f = parse_loadname(srcname);
	load_mem(cm, f.name, f.la, f.offs, f.len, f.mode, 0, 0);

	check(cm);

	destroy_mem(cm);
	exit(0);
    }

    cbm_mem_t *cm = create_mem();

    if (verbose_g) {
	if (opt.load == 0x0801) {
	    printf("jump address $%04X (or RUN).\n", 0x0815);
	} else {
	    printf("jump address $%04X.\n", opt.load);
	}
	printf("using seed 0x%08x.\n", opt.seed);
    }

    generate(cm, opt.rng.sa, opt.rng.ea, opt.load, opt.seed);


    save_mem(cm, opt.outname, opt.rng.sa, opt.rng.ea, 0, MODE_NORMAL);

    destroy_mem(cm);

    exit(0);
}



typedef struct {
    uint8_t a;
    uint8_t b;
} fl16_t;

static const fl16_t s_iv = { 7, 0 };

static uint8_t add1(uint8_t a, uint8_t b)
{
    int c = a + b;
    return (c & 0xff) + (c >> 8);
}

static fl16_t fl16sum(fl16_t s, uint8_t v)
{
    s.a = add1(s.a, v);
    s.b = add1(s.b, s.a);
    return s;
}

static fl16_t fl16range(cbm_mem_t *cm, cbm_ptr_t sa, cbm_ptr_t ea, fl16_t s)
{
    cbm_ptr_t ad = sa;

    while (ad < ea) {
	uint8_t d = get_byte(cm, ad);
	s = fl16sum(s, d);
	ad++;
    }

    return s;
}

static int fl16cmp(fl16_t s1, fl16_t s2)
{
    return (s1.a == s2.a && s1.b == s2.b);
}

static fl16_t fl16random(void)
{
    fl16_t s;
    s.a = random();
    s.b = random();
    if (s.a == 0) {
	s.a = 0xff;
    }
    if (s.b == 0) {
	s.b = 0xff;
    }
    return s;
}


static void gen_random_range(cbm_mem_t *cm, cbm_ptr_t sa, cbm_ptr_t ea, int mode)
{
    cbm_ptr_t ad = sa;
    int n = 0;
    uint8_t v = 0;
    int f = 0;
    while (ad < ea) {
	if (n == 0) {
	    n = (random() & 0xff) + 1;
	    f = random() & 1;
	    v = random();
	}
	if (f) {
	    v = random();
	}
	set_byte(cm, ad, v);
	ad++;
	n--;
    }
}

static int gen_collision(cbm_mem_t *cm, cbm_ptr_t sa, cbm_ptr_t ea, cbm_ptr_t ca, fl16_t my_s)
{
    fl16_t s, init_s;
    init_s = fl16range(cm, sa, ca, s_iv);

    /* find collision */
    int i, j, k;
    for (i = 0x00; i < 0x100; i++) {
	set_byte(cm, ca+0, i);
	for (j = 0x00; j < 0x100; j++) {
	    set_byte(cm, ca+5, j);
	    for (k = 0x00; k < 0x100; k++) {
		set_byte(cm, ca+8, k);

		s = fl16range(cm, ca, ea, init_s);
		if (fl16cmp(s, my_s)) {
		    return 1;
		}
	    }
	}
    }

    return 0;
}

static void generate(cbm_mem_t *cm, cbm_ptr_t sa, cbm_ptr_t ea, cbm_ptr_t load, unsigned int seed)
{
    FixStruct *fs = fs_default_sys;
    cbm_ptr_t ad = load;
    int rm = 0;

    if (load != 0x0801) {
	fs = fs_default_sysless;
    }

#ifdef HAVE_SELF_SUM
    /* find fields */
    cbm_ptr_t sum_start = CBM_PTR_UNDEFINED;
#if 0
    cbm_ptr_t sum_end = CBM_PTR_UNDEFINED;
#endif
    cbm_ptr_t sum_adjust = CBM_PTR_UNDEFINED;
    {
	FixEntry *fe = fs->fix_entry;
	int reloc_offs = -fs->addr + load;
	while (fe->type != ftEnd) {
	    cbm_ptr_t addr = fe->addr + reloc_offs;
	    switch (fe->type) {
	    case ftSelfSumStart:
		sum_start = addr;
		break;
#if 0
	    case ftSelfSumEnd:
		sum_end = addr;
		break;
#endif
	    case ftSelfSumAdjust:
		sum_adjust = addr;
		break;
	    default:
		break;
	    }
	    fe++;
	}
    }
#endif /* HAVE_SELF_SUM */

    srandom(seed);

    insert_mem(fs->data, fs->len, cm, ad);
    ad += fs->len;


    /******
     * Generate CGEN header
     */
    /*
     * HEADER:
     *   00: "CGEN0"
     *   05: <chklen>     ; length of data to checksum
     *   06: <chksum>     ; fletcher-16 sum of data starting adter the chksum
     *                    ; field
     *   08: <header_len> ; length of header from the start of the CGEN0 field
     *   0a: <tag_addr>   ; memory address of the CGEN0 field
     *   0b: ...          ; start of checksum areas
     *
     * ENTRIES:
     *   00: <start addr>
     *   02: <end addr>
     *   04: <checksum>   ; fletcher-16 sum from <start addr> to <end addr>
     *                    ; (exclusive range)
     *   06: <flags>      ; MSB=1 means that this is the last entry
     */
    cbm_ptr_t tag_addr = ad;
    set_byte(cm, ad++, 'C');
    set_byte(cm, ad++, 'G');
    set_byte(cm, ad++, 'E');
    set_byte(cm, ad++, 'N');
    set_byte(cm, ad++, '0');
    set_byte(cm, ad, 0);   /* len to checksum */
    ad++;
    set_word(cm, ad, 0);   /* check sum */
    ad += 2;
    cbm_ptr_t chkstart_addr = ad;
    set_byte(cm, ad, 11);   /* header len */
    ad++;
    set_word(cm, ad, tag_addr);
    ad += 2;

    cbm_ptr_t list_addr = ad;
    set_word(cm, ad, sa);
    ad += 2;
    set_word(cm, ad, ea);
    ad += 2;
    fl16_t my_s = fl16random();
    set_byte(cm, ad, my_s.a);
    ad++;
    set_byte(cm, ad, my_s.b);
    ad++;
    set_byte(cm, ad, 0x80);
    ad++;


    int list_len = ad - chkstart_addr;
    cbm_ptr_t ta = tag_addr;
    ta += 5;
    set_byte(cm, ta, list_len);
    ta++;

    fl16_t s_hd = fl16range(cm, chkstart_addr, ad, s_iv);
    set_byte(cm, ta++, s_hd.a);
    set_byte(cm, ta, s_hd.b);

#ifdef HAVE_SELF_SUM
    uint8_t self_sum = 0x12;
    int sum_len = ad - sum_start;
#endif

    /* process fields to fill in */
    FixEntry *fe = fs->fix_entry;
    int reloc_offs = -fs->addr + load;
    while (fe->type != ftEnd) {
#ifdef HAVE_SELF_SUM
	uint16_t val;
#endif /* HAVE_SELF_SUM */
	cbm_ptr_t addr = fe->addr + reloc_offs;
	switch (fe->type) {
#ifdef HAVE_SELF_SUM
	case ftRelocAdj:
	    val = get_word(cm, addr);
	    val += reloc_offs;
	    set_word(cm, addr, val);
	    break;
#endif /* HAVE_SELF_SUM */
	case ftCheckListPtr:
	    set_word(cm, addr, list_addr);
	    break;
#ifdef HAVE_SELF_SUM
	case ftSelfSumLen:
	    set_byte(cm, addr, sum_len);
	    break;
	case ftSelfSum:
	    set_byte(cm, addr, self_sum);
	    break;
#endif /* HAVE_SELF_SUM */
	default:
	    break;
	}
	fe++;
    }

#ifdef HAVE_SELF_SUM
    /******
     * Calculate and verify self sum
     */
    uint8_t sum = ~self_sum;
    int i;
    for (i = 0; i < sum_len; i++) {
        sum = add1(sum, get_byte(cm, sum_start + i));
    }
    set_byte(cm, sum_adjust, -sum);

    /* verify */
    int c;
    c = 0;
    sum = 0x00;
    for (i = sum_len - 1; i >= 0; --i) {
	int tmp;
        tmp = sum + get_byte(cm, sum_start + i) + c;
	sum = tmp;
	c = tmp & 0x100 ? 1 : 0;
    }
    if (sum != self_sum) {
	panic("couldn't generate self sum! (got 0x%02x, expected 0x%02x)", sum, self_sum);
    }
#endif /* HAVE_SELF_SUM */

    /******
     * Generate random data areas
     */
    if (sa < load) {
	gen_random_range(cm, sa, load, rm);
    }

    gen_random_range(cm, ad, ea, rm);

    /* setup a fake $00 ddr value if $00 isn't included in the range */
    if (sa > 0x00) {
	set_byte(cm, 0x00, 0x2f);
    }
    if (sa < 0x02) {
	while (1) {
	    uint8_t ddr_mask = 0xef;    /* be nice to the tape sense line */

	    /* make sure $00/$01 contains working values */
	    uint8_t ddr = get_byte(cm, 0x00);
	    uint8_t iop = get_byte(cm, 0x01);

	    ddr &= ddr_mask;
	    iop &= 0x3f;
	    iop |= ~ddr_mask;

	    set_byte(cm, 0x00, ddr);
	    set_byte(cm, 0x01, iop);

	    /* check that I/O is visible */
	    iop |= ~ddr;
	    iop &= 0x7;
	    if (iop == 0x05 || iop == 0x06 || iop == 0x07) {
		break;
	    }

	    /* not ok, regenerate */
	    gen_random_range(cm, sa, 0x02, rm);
	}
    }

#define COLLISION_DIST 128
    gen_collision(cm, sa, ea, ea - COLLISION_DIST, my_s);

    /* verify */
    fl16_t vs = fl16range(cm, sa, ea, s_iv);
    if (!fl16cmp(vs, my_s)) {
	panic("failed to find collision!");
    }

}


typedef struct {
    cbm_ptr_t sa;
    cbm_ptr_t ea;
    fl16_t sum;
} sum_entry_t;

#define MAX_NUM_SUM_ENTRIES 64
int num_sum_entries;
sum_entry_t sum_entry[MAX_NUM_SUM_ENTRIES];

static void check(cbm_mem_t *cm)
{
    cbm_ptr_t ad;
    int n = 0;
    char *tag = "CGEN0";
    for (ad = 0; ad < cm->size; ad++) {
	uint8_t d = get_byte(cm, ad);
	if (d == tag[n]) {
	    n++;
	    if (n == 5) {
		break;
	    }
	} else {
	    n = 0;
	}
    }

    if (n != 5) {
	panic("couldn't find tag");
    }
    ad++;

    uint8_t len = get_byte(cm, ad);
    ad++;
    fl16_t s_hd_ref;
    s_hd_ref.a = get_byte(cm, ad);
    ad++;
    s_hd_ref.b = get_byte(cm, ad);
    ad++;
    fl16_t s_hd = fl16range(cm, ad, ad + len, s_iv);

    uint8_t hdlen = get_byte(cm, ad);
    ad++;
    cbm_ptr_t la = get_word(cm, ad) + hdlen;
    ad += 2;

    if (!fl16cmp(s_hd, s_hd_ref)) {
	panic("tag checksum mismatch (got $%02X%02X, expected $%02X%02X)", s_hd.a, s_hd.b, s_hd_ref.a, s_hd_ref.b);
    }

    int end = 0;
    num_sum_entries = 0;
    do {
	sum_entry[num_sum_entries].sa = get_word(cm, la);
	sum_entry[num_sum_entries].ea = get_word(cm, la + 2);
	sum_entry[num_sum_entries].sum.a = get_byte(cm, la + 4);
	sum_entry[num_sum_entries].sum.b = get_byte(cm, la + 5);
	num_sum_entries++;
	end = get_byte(cm, la + 6) & 0x80;
	la += 7;
    } while ( !end );

    int i;
    for (i = 0; i < num_sum_entries; i++) {
	cbm_ptr_t sa = sum_entry[i].sa;
	cbm_ptr_t ea = sum_entry[i].ea;
	if (ea == 0x0000) {
	    ea = 0x10000;
	}

	fl16_t s_ref = sum_entry[i].sum;
	fl16_t s = fl16range(cm, sa, ea, s_iv);

	if (verbose_g) {
	    printf("%d: $%04X-$%04X s=$%02X%02X ($%02X%02X)\n", i, sa, ea, s_ref.a, s_ref.b, s.a, s.b);
	}
	if (!fl16cmp(s, s_ref)) {
	    panic("checksum mismatch in segment $%04X-$%04X (got $%02X%02X, expected $%02X%02X)", sa, ea, s.a, s.b, s_ref.a, s_ref.b);
	}
    }
}

/* eof */
