/**************************************************************************
 *
 * FILE  chd.c
 * Copyright (c) 2007, 2012, 2013, 2018 Daniel Kahlin
 * Written by Daniel Kahlin <daniel@kahlin.net>
 *
 * DESCRIPTION
 *   'hd' like dump for cbm files
 *
 ******/
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "cbm_mem.h"
#include "global.h"
#include "output.h"
#include "params.h"

#define PROGRAM "chd"

const char program_g[] = PROGRAM;

typedef struct {
    cbm_ptr_t overload_addr;
    int longaddr;
    cbm_ptr_t loadaddr;
    int rawfile;
} Options;

static Options opt;

static void dump_file(FILE *fp);

int main(int argc, char *argv[])
{
    int c;

    /* defaults */
    verbose_g = 0;
    debug_g = 0;
    opt.overload_addr = CBM_PTR_UNDEFINED;
    opt.longaddr = 0;
    opt.loadaddr = CBM_PTR_UNDEFINED;
    opt.rawfile = 0;

    /*
     * scan for valid options
     */
    while ( c = getopt(argc, argv, "bl:LO:dvVh"), c != EOF ) {
        switch (c) {
	
	/* a missing parameter */
	case ':':
	/* an illegal option */
	case '?':
	    exit(1);

	/* set verbose */
	case 'v':
	    verbose_g = 1;
	    break;

	/* set debug */
	case 'd':
	    debug_g = 1;
	    break;

	/* print version */
	case 'V':
	    fprintf(stdout, PROGRAM " (" PACKAGE ") " VERSION "\n");
	    exit(0);

	/* print help */
	case 'h':
	    fprintf(stderr,
PROGRAM " (" PACKAGE ") " VERSION ": cbm binary hexdump\n"
"Copyright (c) 2007, 2012, 2013, 2018 Daniel Kahlin <daniel@kahlin.net>\n"
"Written by Daniel Kahlin <daniel@kahlin.net>\n"
"\n"
"usage: " PROGRAM " [-b][-l<addr>][-L][-O[d|t|<addr>][-v][-d][-h][-V] <file>...\n"
"\n"
"Valid options:\n"
"    -b              file has no load address\n"
"    -l              override load address\n"
"    -L              24-bit addressing (default: 16-bit)\n"
"    -O              consider pointer overloading\n"
"    -v              be verbose\n"
"    -d              debugging output\n"
"    -h              displays this help text\n"
"    -V              output program version\n"
	    );
	    exit(0);

	/* set raw mode */
	case 'b':
	    opt.rawfile = 1;
	    break;

	/* set Load address */
	case 'l':
	    opt.loadaddr = parse_value(optarg);
	    break;

	/* set Long addressing */
	case 'L':
	    opt.longaddr = 1;
	    break;

	/* set Overloading */
	case 'O':
	    switch (optarg[0]) {
	    case 't':
		opt.overload_addr = 0xac;
		break;
	    case 'd':
		opt.overload_addr = 0xae;
		break;
	    default:
		opt.overload_addr = parse_value(optarg);
		break;
	    }
	    break;

	/* default behavior */
	default:
	    break;
	}
    }

    /*
     * optind now points at the first non option argument
     */
    int num_files = argc - optind;
    char **srcnames = &argv[optind];

    /* if no arguments, dump from stdin */
    if (num_files == 0) {
	static char *stdinname = "-";
	srcnames = &stdinname;
	num_files++;
    }

    /* if one or more arguments, dump from the listed files */
    int i;
    for (i = 0; i < num_files; i++) {
	int c;
	FILE *fp;
	char *srcname;
	char *print_name;

	srcname = srcnames[i];
	if (!strcmp(srcname, "-")) {
	    print_name = "<stdin>";
	    fp = stdin;
	} else {
	    print_name = srcname;
	    fp = fopen(srcname, "rb");
	    if (!fp) {
		panic("%s: %s", print_name, strerror(errno));
	    }
	}

	if (num_files > 1) {
	    printf("--- %s ---\n", print_name);
	}

	dump_file(fp);

	if (ferror(fp)) {
	    panic("%s: %s", print_name, strerror(errno));
	}

	if (fp != stdin) {
	    c = fclose(fp);
	    if (c == EOF) {
		panic("%s: %s", print_name, strerror(errno));
	    }
	}
    }

    exit(0);
}


static cbm_ptr_t handle_overload(cbm_ptr_t ad, uint8_t c)
{
    cbm_ptr_t next_ad = ad;

    if (is_defined_addr(opt.overload_addr)) {
	if (ad == opt.overload_addr) {
	    next_ad = (ad & 0xffff00) | c;
	}
	if (ad == opt.overload_addr + 1) {
	    next_ad = (ad & 0xff00ff) | (c << 8);
	}
    }

    return next_ad;
}


static cbm_ptr_t inc_addr(cbm_ptr_t ad)
{
    ad++;
    if (!opt.longaddr) {
	ad &= 0xffff;
    } else {
	ad &= 0xffffff;
    }

    return ad;
}



static void dump_file(FILE *fp)
{
    cbm_ptr_t ad;
    cbm_range_t rng = CBM_RANGE_UNDEFINED;
    int nread = 0;
    Dumper dmp;

    init_dumper(&dmp);
    dmp.long_addr = opt.longaddr;

    if (!opt.rawfile) {
	ad = fgetc(fp) + (fgetc(fp) << 8);
	nread += 2;
    } else {
	ad = 0x0000;
    }
    if (is_defined_addr(opt.loadaddr)) {
	ad = opt.loadaddr;
    }

    int c;
    while (c = fgetc(fp), c != EOF) {
	nread++;

	dump_data(&dmp, ad, c);
	update_range(&rng, ad);

	/* check for overloading */
	cbm_ptr_t last_ad = ad;
	ad = handle_overload(ad, c);

	/* if the address made a jump, flush to start a new line */
	if (ad != last_ad) {
	    flush_dumper(&dmp);
	}

	/* next address */
	ad = inc_addr(ad);

    }

    flush_dumper(&dmp);


    int nblks = (nread + 253) / 254;
    if (opt.longaddr) {
	rng.sa &= 0xffffff;
	rng.ea &= 0xffffff;
	printf("dumped $%06X-$%06X (%d %s, %d %s)\n", rng.sa, rng.ea, nread, nread==1?"byte":"bytes", nblks, nblks==1?"block":"blocks");
    } else {
	rng.sa &= 0xffff;
	rng.ea &= 0xffff;
	printf("dumped $%04X-$%04X (%d %s, %d %s)\n", rng.sa, rng.ea, nread, nread==1?"byte":"bytes", nblks, nblks==1?"block":"blocks");
    }
}


/* eof */
