/**************************************************************************
 *
 * FILE  cld.c
 * Copyright (c) 2007, 2012, 2018 Daniel Kahlin
 * Written by Daniel Kahlin <daniel@kahlin.net>
 *
 * DESCRIPTION
 *   cbm file linker
 *
 ******/
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "cbm_mem.h"
#include "global.h"
#include "params.h"
#include "rle.h"

#define PROGRAM "cld"

const char program_g[] = PROGRAM;

#define MAX_NUM_OUTNAMES 256
#define MAX_NUM_RANGES 256

typedef struct {
    char *outname[MAX_NUM_OUTNAMES];
    int num_outnames;
    char *range[MAX_NUM_RANGES];
    int num_ranges;
    int make_exe;
    char *exe_opts;
} Options;

static Options opt;

int main(int argc, char *argv[])
{
    int c;

    /* defaults */
    verbose_g = 0;
    debug_g = 0;
    opt.num_outnames = 0;
    opt.num_ranges = 0;
    opt.make_exe = 0;
    opt.exe_opts = "";

    /*
     * scan for valid options
     */
    while ( c = getopt(argc, argv, "r:xX:o:dvVh"), c != EOF ) {
        switch (c) {

	/* a missing parameter */
	case ':':
	/* an illegal option */
	case '?':
	    exit(1);

	/* set verbose */
	case 'v':
	    verbose_g = 1;
	    break;

	/* set debug */
	case 'd':
	    debug_g = 1;
	    break;

	/* print version */
	case 'V':
	    fprintf(stdout, PROGRAM " (" PACKAGE ") " VERSION "\n");
	    exit(0);

	/* print help */
	case 'h':
	    fprintf(stderr,
PROGRAM " (" PACKAGE ") " VERSION ": cbm binary linker\n"
"Copyright (c) 2007, 2012, 2018 Daniel Kahlin <daniel@kahlin.net>\n"
"Written by Daniel Kahlin <daniel@kahlin.net>\n"
"\n"
"usage: " PROGRAM " [-r<range>][-x][-X<opts>][-o<name>][-v][-d][-h][-V] <source>...\n"
"\n"
"Valid options:\n"
"    -r              specify slice range (may be multiple)\n"
"    -x              produce executable output (implies '-x')\n"
"    -X <opts>       options for executable\n"
"    -o              output file (may be multiple)\n"
"    -v              be verbose\n"
"    -d              debugging output\n"
"    -h              displays this help text\n"
"    -V              output program version\n"
	    );
	    exit(0);

	/* specify save range */
	case 'r':
	    if (opt.num_ranges >= MAX_NUM_RANGES) {
		panic("too many ranges specified");
	    }
	    opt.range[opt.num_ranges++] = optarg;
	    break;

	/* produce executable output */
	case 'x':
	    opt.make_exe = 1;
	    break;

	/* options for executable */
	case 'X':
	    opt.make_exe = 1;
	    opt.exe_opts = optarg;
	    break;

	/* set output file */
	case 'o':
	    if (opt.num_outnames >= MAX_NUM_OUTNAMES) {
		panic("too many outnames specified");
	    }
	    opt.outname[opt.num_outnames++] = optarg;
	    break;

	/* default behavior */
	default:
	    break;
	}
    }

    /*
     * optind now points at the first non option argument
     */
    int num_files = argc - optind;
    char **srcnames = &argv[optind];

    /* if no arguments, dump from stdin */
    if (num_files == 0) {
	static char *stdinname = "-";
	srcnames = &stdinname;
	num_files++;
    }

    if (opt.num_outnames == 0) {
	opt.num_outnames = 1;
	opt.outname[0] = "a.out";
    }


    cbm_mem_t *cm = create_mem();
    int i;

    for (i = 0; i < num_files; i++) {
	char *name = srcnames[i];
	loadfile_t f;

	f = parse_loadname(name);
	load_mem(cm, f.name, f.la, f.offs, f.len, f.mode, 0, 0);
    }


    if (opt.num_ranges) {
	cbm_mem_t *tcm = create_mem();

	for (i = 0; i < opt.num_ranges; i++) {
	    cbm_range_t rng = get_range(cm);
	    slice_t sl;

	    sl = parse_slice(opt.range[i], rng);
	    copy_mem2(cm, sl.sa, sl.ea, tcm, sl.da);
	}

	destroy_mem(cm);
	cm = tcm;
    }

    if (opt.make_exe) {
	cbm_mem_t *dcm = create_mem();
	cbm_range_t rng = get_range(cm);

	rle_pack(dcm, cm, rng.sa, rng.ea, opt.exe_opts);
	destroy_mem(cm);
	cm = dcm;
    }

    for (i = 0; i < opt.num_outnames; i++) {
	char *name = opt.outname[i];
	cbm_range_t rng = get_range(cm);
	savefile_t f;

	f = parse_savename(name, rng);

	save_mem(cm, f.name, f.sa, f.ea, f.la, f.mode);
    }

    destroy_mem(cm);

    exit(0);
}

/* eof */
