/**************************************************************************
 *
 * FILE  cpch.c
 * Copyright (c) 2007, 2012, 2013, 2018 Daniel Kahlin
 * Written by Daniel Kahlin <daniel@kahlin.net>
 *
 * DESCRIPTION
 *   cbm file patcher
 *
 ******/
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "cbm_mem.h"
#include "global.h"
#include "params.h"

#define PROGRAM "cpch"

const char program_g[] = PROGRAM;

#define MAX_NUM_PATCHNAMES 256

typedef struct {
    char *outname;
    char *save_range;
    char *patchname[MAX_NUM_PATCHNAMES];
    int num_patchnames;
    int merge;
} Options;

static Options opt;

void patch_ram(cbm_mem_t *cm, const char *name);
void write_patch(cbm_mem_t *cm, const char *name);

int main(int argc, char *argv[])
{
    int c;

    /* defaults */
    verbose_g = 0;
    debug_g = 0;
    opt.outname = "a.out";
    opt.save_range = "";
    opt.num_patchnames = 0;
    opt.merge = 0;

    /*
     * scan for valid options
     */
    while ( c = getopt(argc, argv, "r:p:o:mdvVh"), c != EOF ) {
        switch (c) {

	/* a missing parameter */
	case ':':
	/* an illegal option */
	case '?':
	    exit(1);

	/* set verbose */
	case 'v':
	    verbose_g = 1;
	    break;

	/* set debug */
	case 'd':
	    debug_g = 1;
	    break;

	/* print version */
	case 'V':
	    fprintf(stdout, PROGRAM " (" PACKAGE ") " VERSION "\n");
	    exit(0);

	/* print help */
	case 'h':
	    fprintf(stderr,
PROGRAM " (" PACKAGE ") " VERSION ": cbm binary patcher\n"
"Copyright (c) 2007, 2012, 2013, 2018 Daniel Kahlin <daniel@kahlin.net>\n"
"Written by Daniel Kahlin <daniel@kahlin.net>\n"
"\n"
"usage: " PROGRAM " [-r<range>][-p<patch>][-o<name>][-m][-v][-d][-h][-V] <source>...\n"
"\n"
"Valid options:\n"
"    -r              specify save range\n"
"    -p              patch file\n"
"    -o              output file\n"
"    -m              merge patch files\n"
"    -v              be verbose\n"
"    -d              debugging output\n"
"    -h              displays this help text\n"
"    -V              output program version\n"
	    );
	    exit(0);

	/* specify save range */
	case 'r':
	    opt.save_range = optarg;
	    break;

	/* set output file */
	case 'o':
	    opt.outname = optarg;
	    break;

	/* set patchname file */
	case 'p':
	    if (opt.num_patchnames >= MAX_NUM_PATCHNAMES) {
		panic("too many patchnames specified");
	    }
	    opt.patchname[opt.num_patchnames++] = optarg;
	    break;

	/* set merge mode */
	case 'm':
	    opt.merge = 1;
	    break;

	/* default behavior */
	default:
	    break;
	}
    }

    /*
     * optind now points at the first non option argument
     */
    int num_files = argc - optind;
    char *srcname = "";

    if (!opt.merge) {
	if (num_files < 1) {
	    panic("too few arguments");
	}
	srcname = argv[optind];
    } else {
	if (num_files) {
	    panic("too many arguments");
	}
    }

    cbm_mem_t *cm = create_mem();

    if (!opt.merge) {
	loadfile_t f;

	f = parse_loadname(srcname);
	load_mem(cm, f.name, f.la, f.offs, f.len, f.mode, 0, 0);
    }

    /* read patch files */
    int i;
    for (i = 0; i < opt.num_patchnames; i++) {
	patch_ram(cm, opt.patchname[i]);
    }

    if (!opt.merge) {
	cbm_range_t rng = get_range(cm);
	parse_range(opt.save_range, &rng);
	/* write destination file */
	save_mem(cm, opt.outname, rng.sa, rng.ea, 0, MODE_NORMAL);
    } else {
	write_patch(cm, opt.outname);
    }

    destroy_mem(cm);
    exit(0);
}


void patch_ram(cbm_mem_t *cm, const char *name)
{
    int c;
    FILE *fp;
    int len;
    uint8_t buf[256];
    uint8_t *bp;
    int num;
    int cnt;

    fp = fopen(name, "rb");
    if (!fp) {
	panic("%s: %s", name, strerror(errno));
    }

    if (verbose_g) {
	printf("patch file '%s':\n", name);
    }
    /*
     * parse header
     */

    /* skip start addr */
    fgetc(fp);
    fgetc(fp);
    /* read header */
    len = fgetc(fp) | fgetc(fp) << 8;
    size_t n = fread(buf, 1, 4, fp);
    if (n != 4) {
	panic("couldn't read header");
    }
    bp = buf;

    /* check signature */
    if (len >= 4 && strncmp((char *)bp, "PTCH", 4)) {
	panic("not a patch file");
    }
    bp += 4;
    len -= 4;


    /*
     * parse chunks
     */
    num = 0;
    cnt = 0;
    while (1) {
	int i;
	cbm_ptr_t addr;
	int len;
	addr = fgetc(fp) | fgetc(fp) << 8;
	len = fgetc(fp) | fgetc(fp) << 8;
	if ( len == 0 || feof(fp) || ferror(fp) ) {
	    break;
	}
	for (i = 0; i < len; i++) {
	    uint8_t c = fgetc(fp);
	    set_byte(cm, addr + i, c);
	}

	if (verbose_g) {
	    printf("  chunk %d: $%04x-$%04x (%d bytes)\n", num, addr, addr + len - 1, len);
	}

	cnt += len;
	num++;
    }

    if (ferror(fp)) {
	panic("%s: %s", name, strerror(errno));
    }

    c = fclose(fp);
    if (c == EOF) {
	panic("%s: %s", name, strerror(errno));
    }
    if (verbose_g) {
	printf("  %d patch bytes in %d chunks.\n", cnt, num);
    }
}


void write_patch(cbm_mem_t *cm, const char *name)
{
    int c;
    FILE *fp;
    int num;
    int cnt;
    int len;
    cbm_range_t rng = get_range(cm);

    fp = fopen(name, "wb");
    if (!fp) {
	panic("%s: %s", name, strerror(errno));
    }

    if (verbose_g) {
	printf("patch file '%s':\n", name);
    }
    /*
     * write header
     */

    /* skip start addr */
    fputc(0, fp);
    fputc(0, fp);
    len = 4;
    fputc(len & 0xff, fp);
    fputc(len >> 8, fp);
    fputc('P', fp);
    fputc('T', fp);
    fputc('C', fp);
    fputc('H', fp);

    /*
     * write chunks
     */
    num = 0;
    cnt = 0;
    cbm_ptr_t ad = rng.sa;
    len = 0;
    cbm_ptr_t addr = ad;
    while (!ferror(fp)) {
	if (ad < rng.ea && get_byte_defined(cm, ad)) {
	    if (len == 0) {
		addr = ad;
	    }
	    len++;
	} else {
	    if (len > 0 ) {
		int i;
		fputc(addr & 0xff, fp);
		fputc(addr >> 8, fp);
		fputc(len & 0xff, fp);
		fputc(len >> 8, fp);
		for (i = 0; i < len; i++) {
		    fputc(get_byte(cm, addr + i), fp);
		}
		if (verbose_g) {
		    printf("  chunk %d: $%04x-$%04x (%d bytes)\n", num, addr, addr + len - 1, len);
		}
		num++;
		cnt += len;
	    }
	    if (ad >= rng.ea) {
		break;
	    }
	    len = 0;
	}

	ad++;
    }

    /* write end marker */
    fputc(0, fp);
    fputc(0, fp);
    fputc(0, fp);
    fputc(0, fp);

    if (ferror(fp)) {
	panic("%s: %s", name, strerror(errno));
    }

    c = fclose(fp);
    if (c == EOF) {
	panic("%s: %s", name, strerror(errno));
    }
    if (verbose_g) {
	printf("  %d patch bytes in %d chunks.\n", cnt, num);
    }
}


/* eof */
