/**************************************************************************
 *
 * FILE  csiz.c
 * Copyright (c) 2018 Daniel Kahlin
 * Written by Daniel Kahlin <daniel@kahlin.net>
 *
 * DESCRIPTION
 *   cbm file range checker
 *
 ******/
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "global.h"
#include "params.h"

#define PROGRAM "csiz"

const char program_g[] = PROGRAM;

typedef struct {
    int show_start_addr;
    int show_end_addr;
    int inclusive_end_addr;
    int have_0x_format;
    int have_newline;
} Options;

static Options opt;

static void analyze_file(FILE *fp);

int main(int argc, char *argv[])
{
    int c;

    /* defaults */
    verbose_g = 0;
    debug_g = 0;
    opt.show_start_addr = 0;
    opt.show_end_addr = 0;
    opt.inclusive_end_addr = 0;
    opt.have_0x_format = 0;
    opt.have_newline = 1;

    /*
     * scan for valid options
     */
    while ( c = getopt(argc, argv, "seixndvVh"), c != EOF ) {
        switch (c) {

	/* a missing parameter */
	case ':':
	/* an illegal option */
	case '?':
	    exit(1);

	/* set verbose */
	case 'v':
	    verbose_g = 1;
	    break;

	/* set debug */
	case 'd':
	    debug_g = 1;
	    break;

	/* print version */
	case 'V':
	    fprintf(stdout, PROGRAM " (" PACKAGE ") " VERSION "\n");
	    exit(0);

	/* print help */
	case 'h':
	    fprintf(stderr,
PROGRAM " (" PACKAGE ") " VERSION ": cbm binary range checker\n"
"Copyright (c) 2018 Daniel Kahlin <daniel@kahlin.net>\n"
"Written by Daniel Kahlin <daniel@kahlin.net>\n"
"\n"
"usage: " PROGRAM " [-s][-e][-i][-x][-n][-v][-d][-h][-V] <file>...\n"
"\n"
"Valid options:\n"
"    -s              start address\n"
"    -e              end address\n"
"    -i              inclusive end address\n"
"    -x              0x format\n"
"    -n              do not output trailing newline\n"
"    -v              be verbose\n"
"    -d              debugging output\n"
"    -h              displays this help text\n"
"    -V              output program version\n"
	    );
	    exit(0);

	/* show start address */
	case 's':
	    opt.show_start_addr = 1;
	    break;

	/* show end address */
	case 'e':
	    opt.show_end_addr = 1;
	    break;

	/* inclusive end address */
	case 'i':
	    opt.inclusive_end_addr = 1;
	    break;

	/* 0x format */
	case 'x':
	    opt.have_0x_format = 1;
	    break;

	/* disable trailing newline */
	case 'n':
	    opt.have_newline = 0;
	    break;

	/* default behavior */
	default:
	    break;
	}
    }

    /*
     * optind now points at the first non option argument
     */
    int num_files = argc - optind;
    char **srcnames = &argv[optind];

    /* if no arguments, dump from stdin */
    if (num_files == 0) {
	static char *stdinname = "-";
	srcnames = &stdinname;
	num_files++;
    }

    /* if one or more arguments, dump from the listed files */
    int i;
    for (i = 0; i < num_files; i++) {
	int c;
	FILE *fp;
	char *srcname;
	char *print_name;

	srcname = srcnames[i];
	if (!strcmp(srcname, "-")) {
	    print_name = "<stdin>";
	    fp = stdin;
	} else {
	    print_name = srcname;
	    fp = fopen(srcname, "rb");
	    if (!fp) {
		panic("%s: %s", print_name, strerror(errno));
	    }
	}

	if (num_files > 1) {
	    printf("%s: ", print_name);
	}

	analyze_file(fp);

	if (ferror(fp)) {
	    panic("%s: %s", print_name, strerror(errno));
	}

	if (opt.have_newline || i < (num_files - 1)) {
	    putchar('\n');
	}

	if (fp != stdin) {
	    c = fclose(fp);
	    if (c == EOF) {
		panic("%s: %s", print_name, strerror(errno));
	    }
	}
    }
    exit(0);
}


static void print_addr(cbm_ptr_t ad)
{
    if (opt.have_0x_format) {
	printf("0x");
	printf("%04x", ad & 0xffff);
    } else {
	printf("$");
	printf("%04X", ad & 0xffff);
    }
}
static void print_start_addr(cbm_ptr_t sa)
{
    print_addr(sa);
}
static void print_end_addr(cbm_ptr_t ea)
{
    if (opt.inclusive_end_addr) {
	print_addr(ea - 1);
    } else {
	print_addr(ea);
    }
}

static void analyze_file(FILE *fp)
{
    size_t n;
    uint16_t sa, ea;

    sa = fgetc(fp) & 0xff;
    sa |= (fgetc(fp) << 8) & 0xff00;

    n = 2;
    int c;
    while (c = fgetc(fp), c != EOF) {
	n++;
    }

    ea = sa + n - 2;

    if (ferror(fp)) {
	return;
    }

    int output_range = 0;
    int output_length = 0;
    int output_start = 0;
    int output_end = 0;

    if (opt.show_start_addr == 0 && opt.show_end_addr == 0) {
	output_range = 1;
	output_length = 1;
    } else if (opt.show_start_addr == 1 && opt.show_end_addr == 1) {
	output_range = 1;
    } else {
	output_start = opt.show_start_addr;
	output_end = opt.show_end_addr;
    }

    if (output_range) {
	print_start_addr(sa);
	if (!opt.inclusive_end_addr) {
	    putchar('-');
	} else {
	    putchar('+');
	}
	print_end_addr(ea);
    }
    if (output_start) {
	print_start_addr(sa);
    }
    if (output_end) {
	print_end_addr(ea);
    }
    if (output_length) {
	int n_blks = (n + 253) / 254;
	printf(" (%u %s, %u %s)", (unsigned)n, n == 1 ? "byte" : "bytes", (unsigned)n_blks, n_blks == 1 ? "block" : "blocks");
    }
}

/* eof */
