/**************************************************************************
 *
 * FILE  params.c
 * Copyright (c) 2012, 2018 Daniel Kahlin
 * Written by Daniel Kahlin <daniel@kahlin.net>
 *
 * DESCRIPTION
 *   parameter parsing
 *
 ******/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cbm_mem.h"
#include "global.h"
#include "params.h"



static unsigned long int my_strtoul(const char *ptr, char **eptr)
{
    int base;
    unsigned long int v;

    switch (*ptr) {
    case '$':
	base = 16;
	ptr++;
	break;
    case '@':
	base = 8;
	ptr++;
	break;
    case '%':
	base = 2;
	ptr++;
	break;
    default:
	base = 0;
	break;
    }

    v = strtoul(ptr, eptr, base);

    return v;
}


static char conv_trail(char c)
{
    if (c == 0) {
	return '\n';
    }
    return c;
}

#define FLAG_MAY_BE_EMPTY (1<<0)
static int parse_ptr_core(const char *ptr, char **eptr, const char *allowed_trail, int flags, cbm_ptr_t *ad)
{
    unsigned long int v;

    if (flags & FLAG_MAY_BE_EMPTY) {
	if ( strchr(allowed_trail, conv_trail(*ptr)) ) {
	    /*
	     * UGLY:
	     * this cast seems necessary to have the strtoul style interface
	     */
	    *eptr = (char *)ptr;
	    return 0;
	}
    }

    v = my_strtoul(ptr, eptr);

    if (ptr == *eptr) {
	return 1;
    }

    if ( strchr(allowed_trail, conv_trail(**eptr)) ) {
	*ad = v;
	return 0;
    }

    return 1;
}


static int parse_range_core(const char *str, char **eptr, const char *allowed_trail, int flags, cbm_range_t *rng)
{
    char c;
    char *p;
    int have_addr = 0;

    if (flags & FLAG_MAY_BE_EMPTY) {
	if ( !(str && strlen(str)) ) {
	    /*
	     * UGLY:
	     * this cast seems necessary to have the strtoul style interface
	     */
	    *eptr = (char *)str;
	    return 0;
	}
    }

    if (parse_ptr_core(str, eptr, "+-", FLAG_MAY_BE_EMPTY, &rng->sa)) {
	return 1;
    }
    if (str != *eptr) {
	have_addr = 1;
    }
    p = *eptr;
    c = *p++;

    if (parse_ptr_core(p, eptr, allowed_trail, FLAG_MAY_BE_EMPTY, &rng->ea)) {
	return 1;
    }
    if (p != *eptr) {
	have_addr = 1;
    }
    if (p != *eptr && c == '+') {
	/* inclusive range */
	rng->ea++;
    }

    if (!have_addr) {
	return 1;
    }

    return 0;
}



int parse_value(char *str)
{
    char *eptr;
    unsigned long int v;

    v = my_strtoul(str, &eptr);
    if ( str == eptr || *eptr != 0 ) {
	panic("couldn't parse value (%s)", str);
    }

    return v;
}

void parse_range(char *str, cbm_range_t *rng)
{
    char *eptr;

    if (parse_range_core(str, &eptr, "\n", FLAG_MAY_BE_EMPTY, rng)) {
	panic("couldn't parse range (%s)", str);
    }
}


slice_t parse_slice(char *str, cbm_range_t rng)
{
    slice_t sl;
    char *p;
    char *eptr;

    if (parse_range_core(str, &eptr, ",\n", FLAG_MAY_BE_EMPTY, &rng)) {
	panic("couldn't parse slice (%s)", str);
    }
    p = eptr;

    sl.sa = rng.sa;
    sl.ea = rng.ea;
    sl.da = rng.sa;

    if (*p == ',') {
	p++;
	if (parse_ptr_core(p, &eptr, "\n", 0, &sl.da)) {
	    panic("couldn't parse slice (%s)", str);
	}
    }

    return sl;
}

loadfile_t parse_loadname(char *name)
{
    char *p, *name_end;
    char c;
    char *eptr;
    unsigned long int v;
    loadfile_t file;

    file.name = name;
    file.mode = MODE_NORMAL;
    file.la = CBM_PTR_UNDEFINED;
    file.offs = -1;
    file.len = -1;


    p = name;
    /* parse */
    p = strpbrk(p, ",@");
    if (p) {
	name_end = p;
	c = *p++;

	switch (c) {
	case ',':
	    v = my_strtoul(p, &eptr);
	    if (p == eptr && *p != ',') {
		panic("missing start address");
	    }
	    if (p != eptr) {
		file.la = v;
		file.mode = MODE_NEWADDR;
	    }

	    p = eptr;
	    break;
	case '@':
	    file.la = my_strtoul(p, &eptr);
	    if (p == eptr) {
		panic("missing start address");
	    }
	    file.mode = MODE_RAW;

	    p = eptr;
	    break;
	default:
	    break;
	}

	p = strchr(p, ',');
	if (p) {
	    p++;
	    v = my_strtoul(p, &eptr);
	    if (p != eptr) {
		file.offs = v;
	    }

	    p = strchr(p, ',');
	    if (p) {
		p++;
		file.len = parse_value(p);
	    }
	}

	/* mark the end of the file name */
	*name_end = 0;
    }

    if (debug_g) {
	printf("name='%s', ad=$%04X, offs=%d, len=%d, mode=%d\n", file.name, file.la, file.offs, file.len, file.mode);
    }

    return file;
}

savefile_t parse_savename(char *name, cbm_range_t rng)
{
    char *p, *name_end;
    char c;
    char *eptr;
    savefile_t file;

    file.name = name;
    file.mode = MODE_NORMAL;
    file.la = CBM_PTR_UNDEFINED;
    file.sa = rng.sa;
    file.ea = rng.ea;


    p = name;
    /* parse */
    p = strpbrk(p, ",@");
    if (p) {
	name_end = p;
	c = *p++;

	switch (c) {
	case ',':
	    if (parse_range_core(p, &eptr, ",@\n", 0, &rng)) {
		if (parse_ptr_core(p, &eptr, "\n", 0, &file.la)) {
		    panic("couldn't parse savename (%s)", name);
		}
		p = eptr;
		file.mode = MODE_NEWADDR;
	    } else {
		file.sa = rng.sa;
		file.ea = rng.ea;
		p = eptr;
		c = *p++;
		switch (c) {
		case ',':
		    if (parse_ptr_core(p, &eptr, "\n", 0, &file.la)) {
			panic("couldn't parse savename (%s)", name);
		    }
		    p = eptr;
		    file.mode = MODE_NEWADDR;
		    break;
		case '@':
		    file.mode = MODE_RAW;
		    break;
		case '\0':
		    p--;
		    break;
		default:
		    break;
		}
	    }
	    break;
	case '@':
	    file.mode = MODE_RAW;
	    break;
	default:
	    break;
	}
	if (*p != 0) {
	    panic("couldn't parse savename (%s)", name);
	}


	/* mark the end of the file name */
	*name_end = 0;
    }

    if (debug_g) {
	printf("name='%s', la=$%04X, sa=$%04X, ea=$%04X, mode=%d\n", file.name, file.la, file.sa, file.ea, file.mode);
    }

    return file;
}

/* eof */
